@@ -16,8 +16,14 @@
 
 #define DEFAULT_CHAR (' ' << 8)
 
-#define MENUITEM_SETFONT 1
-#define MENUITEM_FULLSCREEN 2
+enum {
+	MENUITEM_SETFONT,
+	MENUITEM_FULLSCREEN,
+	MENUITEM_SETBGCOL,
+	MENUITEM_SETDIMCOL,
+	MENUITEM_SETNORMALCOL,
+	MENUITEM_SETBRIGHTCOL,
+};
 
 #define REGISTRY_PATH "Software\\Cowlark Technologies\\WordGrinder"
 
@@ -29,8 +35,6 @@ static int screenwidth = 0;
 static int screenheight = 0;
 static int defaultattr = 0;
 
-static UINT_PTR timer = 0;
-static bool cursor_visible = false;
 static int cursorx = 0;
 static int cursory = 0;
 
@@ -39,11 +43,14 @@ static bool window_geometry_valid = false;
 static RECT window_geometry;
 static bool window_created = false;
 
-static void resize_buffer(void);
+static void resize_buffer(bool force);
 static void fullscreen_cb(void);
 static void switch_to_full_screen(void);
 static void switch_to_windowed(void);
-static void reset_cursor(void);
+
+static COLORREF custom_colours[16];
+
+COLORREF colourmap[COLOUR__NUM];
 
 static HKEY make_key(const char* keystring)
 {
@@ -83,6 +90,33 @@ static void write_default_font(void)
 	RegCloseKey(key);
 }
 
+static void read_colourmap(void)
+{
+	HKEY key = make_key(REGISTRY_PATH);
+	DWORD size = sizeof(colourmap);
+	DWORD e = RegQueryValueEx(key, "Colours",
+			NULL, NULL,
+			(LPBYTE) colourmap, &size);
+	RegCloseKey(key);
+
+	if (e != ERROR_SUCCESS)
+	{
+		colourmap[COLOUR_BLACK]  = 0x000000;
+		colourmap[COLOUR_DIM]    = 0x555555;
+		colourmap[COLOUR_NORMAL] = 0x888888;
+		colourmap[COLOUR_BRIGHT] = 0xffffff;
+	}
+}
+
+static void write_colourmap(void)
+{
+	HKEY key = make_key(REGISTRY_PATH);
+	RegSetValueEx(key, "Colours", 0,
+			REG_BINARY,
+			(LPBYTE) colourmap, sizeof(colourmap));
+	RegCloseKey(key);
+}
+
 static void read_window_geometry(void)
 {
 	HKEY key = make_key(REGISTRY_PATH);
@@ -185,20 +219,23 @@ static void paint_cb(HWND window, PAINTSTRUCT* ps, HDC dc)
 	int textwidth, textheight;
 	glyphcache_getfontsize(&textwidth, &textheight);
 
-	struct glyph* cursorglyph = NULL;
-
 	int x1 = ps->rcPaint.left / textwidth;
 	x1 -= 1; /* because of overlapping characters */
 	if (x1 < 0)
 		x1 = 0;
 
+	HBRUSH blackbrush = CreateSolidBrush(colourmap[COLOUR_BLACK]);
+	HPEN dimpen = CreatePen(PS_SOLID, 0, colourmap[COLOUR_DIM]);
+	HPEN normalpen = CreatePen(PS_SOLID, 0, colourmap[COLOUR_NORMAL]);
+	HPEN brightpen = CreatePen(PS_SOLID, 0, colourmap[COLOUR_BRIGHT]);
+
 	int y1 = ps->rcPaint.top/textheight;
 	int x2 = ps->rcPaint.right/textwidth;
 	x2 += 1; /* because of overlapping characters */
 	if (x2 >= screenwidth)
 	{
 		RECT r = {screenwidth*textwidth, 0, ps->rcPaint.right, ps->rcPaint.bottom};
-		FillRect(dc, &r, GetStockObject(BLACK_BRUSH));
+		FillRect(dc, &r, blackbrush);
 		x2 = screenwidth;
 	}
 
@@ -206,16 +243,12 @@ static void paint_cb(HWND window, PAINTSTRUCT* ps, HDC dc)
 	if (y2 >= screenheight)
 	{
 		RECT r = {0, screenheight*textheight, ps->rcPaint.right, ps->rcPaint.bottom};
-		FillRect(dc, &r, GetStockObject(BLACK_BRUSH));
+		FillRect(dc, &r, blackbrush);
 		y2 = screenheight-1;
 	}
 
 	int state = SaveDC(dc);
 
-	HPEN brightpen = CreatePen(PS_SOLID, 0, 0xffffff);
-	HPEN normalpen = CreatePen(PS_SOLID, 0, 0x888888);
-	HPEN dimpen = CreatePen(PS_SOLID, 0, 0x555555);
-
 	for (int y = y1; y <= y2; y++)
 	{
 		int sy = y * textheight;
@@ -223,7 +256,7 @@ static void paint_cb(HWND window, PAINTSTRUCT* ps, HDC dc)
 		/* Clear this line (or at least the part of it we're drawing). */
 
 		RECT r = {ps->rcPaint.left, sy, ps->rcPaint.right, sy+textheight};
-		FillRect(dc, &r, GetStockObject(BLACK_BRUSH));
+		FillRect(dc, &r, blackbrush);
 
 		/* Draw the actual text. */
 
@@ -253,9 +286,6 @@ static void paint_cb(HWND window, PAINTSTRUCT* ps, HDC dc)
 					LineTo(dc, sx+glyph->width, sy+textheight-1);
 				}
 			}
-
-			if ((x == cursorx) && (y == cursory))
-				cursorglyph = glyph;
 		}
 
 		/* Now go through and invert any characters which are in reverse. */
@@ -280,15 +310,26 @@ static void paint_cb(HWND window, PAINTSTRUCT* ps, HDC dc)
 		}
 	}
 
-	/* Invert the square containing the cursor. */
+	/* Draw the cursor caret. */
+
+	{
+		int x = cursorx*textwidth;
+		int y = cursory*textheight;
+
+		SelectObject(dc, brightpen);
+		MoveToEx(dc, x, y, NULL);
+		LineTo(dc, x, y+textheight);
+		SetPixelV(dc, x-1, y-1, 0xffffff);
+		SetPixelV(dc, x+1, y-1, 0xffffff);
+		SetPixelV(dc, x-1, y+textheight, 0xffffff);
+		SetPixelV(dc, x+1, y+textheight, 0xffffff);
+	}
 
-	if (cursor_visible && cursorglyph)
-		BitBlt(dc, cursorx*textwidth, cursory*textheight, cursorglyph->width, textheight,
-				NULL, 0, 0, DSTINVERT);
 
 	DeleteObject(brightpen);
 	DeleteObject(normalpen);
 	DeleteObject(dimpen);
+	DeleteObject(blackbrush);
 	RestoreDC(dc, state);
 }
 
@@ -311,7 +352,25 @@ static void setfont_cb(void)
 		glyphcache_init(dc, &fontlf);
 		ReleaseDC(window, dc);
 
-		resize_buffer();
+		resize_buffer(false);
+	}
+}
+
+static void setcolour_cb(int c)
+{
+	CHOOSECOLOR cc = {sizeof(cc)};
+	cc.hwndOwner = window;
+	cc.Flags = CC_ANYCOLOR | CC_RGBINIT;
+	cc.rgbResult = colourmap[c];
+	cc.lpCustColors = custom_colours;
+
+	if (ChooseColor(&cc))
+	{
+		colourmap[c] = cc.rgbResult;
+		write_colourmap();
+
+		glyphcache_flush();
+		resize_buffer(true);
 	}
 }
 
@@ -330,13 +389,14 @@ static void fullscreen_cb(void)
 	else
 		switch_to_windowed();
 
-	resize_buffer();
+	resize_buffer(false);
 }
 
 static void create_cb(void)
 {
 	/* Initialise the glyph cache. */
 
+	read_colourmap();
 	read_default_font();
 
 	HDC dc = GetDC(window);
@@ -409,7 +469,7 @@ static LRESULT CALLBACK window_cb(HWND window, UINT message,
 				create_cb();
 				window_created = true;
 			}
-			resize_buffer();
+			resize_buffer(false);
 			break;
 
 		case WM_SIZING:
@@ -441,7 +501,6 @@ static LRESULT CALLBACK window_cb(HWND window, UINT message,
 
 		case WM_CHAR:
 		{
-			reset_cursor();
 			unicode_key(wparam, lparam);
 			break;
 		}
@@ -449,8 +508,6 @@ static LRESULT CALLBACK window_cb(HWND window, UINT message,
 		case WM_KEYDOWN:
 		case WM_SYSKEYDOWN:
 		{
-			reset_cursor();
-
 			if (special_key(wparam, lparam))
 				return 1;
 			break;
@@ -464,6 +521,19 @@ static LRESULT CALLBACK window_cb(HWND window, UINT message,
 					setfont_cb();
 					break;
 
+				case MENUITEM_SETBGCOL:
+					setcolour_cb(COLOUR_BLACK);
+					break;
+				case MENUITEM_SETDIMCOL:
+					setcolour_cb(COLOUR_DIM);
+					break;
+				case MENUITEM_SETNORMALCOL:
+					setcolour_cb(COLOUR_NORMAL);
+					break;
+				case MENUITEM_SETBRIGHTCOL:
+					setcolour_cb(COLOUR_BRIGHT);
+					break;
+
 				case MENUITEM_FULLSCREEN:
 					fullscreen_cb();
 					break;
@@ -502,7 +572,7 @@ void dpy_init(const char* argv[])
 	read_window_geometry();
 }
 
-static void resize_buffer(void)
+static void resize_buffer(bool force)
 {
 	RECT rect;
 	int e = GetClientRect(window, &rect);
@@ -515,7 +585,7 @@ static void resize_buffer(void)
 	int w = rect.right / textwidth;
 	int h = rect.bottom / textheight;
 
-	if ((w != screenwidth) || (h != screenheight))
+	if (force || (w != screenwidth) || (h != screenheight))
 	{
 		glyphcache_flush();
 
@@ -586,32 +656,32 @@ static void switch_to_full_screen(void)
 	ShowWindow(window, SW_SHOWDEFAULT);
 }
 
+/* Actually invalidates a 3x3 square around the character, to deal with
+ * overdraw. */
 static void invalidate_character_at(int x, int y)
 {
 	int textwidth, textheight;
 	glyphcache_getfontsize(&textwidth, &textheight);
 
 	RECT r;
-	r.left = cursorx * textwidth;
-	r.top = cursory * textheight;
-	r.right = r.left + textwidth;
-	r.bottom = r.top + textheight;
+	r.left = cursorx * textwidth - 1;
+	r.top = cursory * textheight - 1;
+	r.right = r.left + textwidth + 2;
+	r.bottom = r.top + textheight + 2;
 	InvalidateRect(window, &r, 0);
 }
 
-static VOID CALLBACK cursor_blink_cb(HWND hwnd, UINT message,
-    UINT_PTR timer, DWORD time)
+static void insert_item_on_menu(const char* msg, int id, HMENU menu)
 {
-	cursor_visible = !cursor_visible;
+	MENUITEMINFO mii = {sizeof(mii)};
+	mii.fMask = MIIM_FTYPE | MIIM_STRING | MIIM_ID;
+	mii.fType = MFT_STRING;
+	mii.dwTypeData = (char*) msg;
+	mii.cch = strlen(msg);
+	mii.wID = id;
 
-	invalidate_character_at(cursorx, cursory);
-	UpdateWindow(window);
-}
-
-static void reset_cursor(void)
-{
-	timer = SetTimer(NULL, timer, 500, cursor_blink_cb);
-	cursor_visible = true;
+	int count = GetMenuItemCount(menu);
+	InsertMenuItem(menu, count+1, TRUE, &mii);
 }
 
 static void switch_to_windowed(void)
@@ -644,28 +714,29 @@ static void switch_to_windowed(void)
 
 	{
 		HMENU menu = GetSystemMenu(window, FALSE);
-		int count = GetMenuItemCount(menu);
 
 		MENUITEMINFO mii;
 		mii.cbSize = sizeof(mii);
 
 		mii.fMask = MIIM_FTYPE;
 		mii.fType = MFT_SEPARATOR;
-		InsertMenuItem(menu, count++, TRUE, &mii);
+		int count = GetMenuItemCount(menu);
+		InsertMenuItem(menu, count+1, TRUE, &mii);
 
-		mii.fMask = MIIM_FTYPE | MIIM_STRING | MIIM_ID;
-		mii.fType = MFT_STRING;
-		mii.dwTypeData = "Select display fon&t...";
-		mii.cch = strlen(mii.dwTypeData);
-		mii.wID = MENUITEM_SETFONT;
-		InsertMenuItem(menu, count++, TRUE, &mii);
+		insert_item_on_menu("Select display fon&t...",
+			MENUITEM_SETFONT, menu);
 
-		mii.fMask = MIIM_FTYPE | MIIM_STRING | MIIM_ID;
-		mii.fType = MFT_STRING;
-		mii.dwTypeData = "&Fullscreen mode\tAlt+Enter";
-		mii.cch = strlen(mii.dwTypeData);
-		mii.wID = MENUITEM_FULLSCREEN;
-		InsertMenuItem(menu, count++, TRUE, &mii);
+		insert_item_on_menu("Select &background colour...",
+			MENUITEM_SETBGCOL, menu);
+		insert_item_on_menu("Select &dim colour...",
+			MENUITEM_SETDIMCOL, menu);
+		insert_item_on_menu("Select &normal colour...",
+			MENUITEM_SETNORMALCOL, menu);
+		insert_item_on_menu("Select b&right colour...",
+			MENUITEM_SETBRIGHTCOL, menu);
+
+		insert_item_on_menu("&Fullscreen mode\tAlt+Enter",
+			MENUITEM_FULLSCREEN, menu);
 
 		EnableMenuItem(menu, SC_CLOSE,
 			MF_BYCOMMAND | MF_DISABLED | MF_GRAYED);
@@ -698,7 +769,6 @@ void dpy_start(void)
 		}
 	}
 
-	reset_cursor();
 	switch_to_windowed();
 }
 
