@@ -9,7 +9,7 @@
 #ifndef GDI_H
 #define GDI_H
 
-#include "uthash.h"
+#include "utils/uthash.h"
 
 #define VKM_SHIFT      0x100
 #define VKM_CTRL       0x200
@@ -43,4 +43,14 @@ extern void dpy_flushkeys(void);
 
 extern HWND window;
 
+enum {
+	COLOUR_BLACK,
+	COLOUR_DIM,
+	COLOUR_NORMAL,
+	COLOUR_BRIGHT,
+	COLOUR__NUM
+};
+
+extern COLORREF colourmap[COLOUR__NUM];
+
 #endif
