@@ -1,9 +1,6 @@
 /* © 2008 David Given.
  * WordGrinder is licensed under the MIT open source license. See the COPYING
  * file in this distribution for the full text.
- *
- * $Id: wcwidth.c 159 2009-12-13 13:11:03Z dtrg $
- * $URL: https://wordgrinder.svn.sf.net/svnroot/wordgrinder/wordgrinder/src/c/emu/wcwidth.c $
  */
 
 #include "globals.h"
@@ -129,7 +126,7 @@ static int bisearch(wchar_t ucs, const struct interval *table, int max) {
  * in ISO 10646.
  */
 
-int wcwidth(uni_t ucs)
+int emu_wcwidth(uni_t ucs)
 {
   /* sorted list of non-overlapping intervals of non-spacing characters */
   /* generated by "uniset +cat=Me +cat=Mn +cat=Cf -00AD +1160-11FF +200B c" */
