@@ -65,6 +65,18 @@ static int setdim_cb(lua_State* L)
 	return 0;
 }
 
+static int setbright_cb(lua_State* L)
+{
+	dpy_setattr(-1, DPY_BRIGHT);
+	return 0;
+}
+
+static int setitalic_cb(lua_State* L)
+{
+	dpy_setattr(-1, DPY_ITALIC);
+	return 0;
+}
+
 static int setnormal_cb(lua_State* L)
 {
 	dpy_setattr(0, 0);
@@ -220,7 +232,9 @@ void screen_init(const char* argv[])
 		{ "setbold",                   setbold_cb },
 		{ "setunderline",              setunderline_cb },
 		{ "setreverse",                setreverse_cb },
+		{ "setbright",                 setbright_cb },
 		{ "setdim",                    setdim_cb },
+		{ "setitalic",                 setitalic_cb },
 		{ "setnormal",                 setnormal_cb },
 		{ "write",                     write_cb },
 		{ "cleararea",                 cleararea_cb },
