@@ -121,7 +121,8 @@ static int writestyled_cb(lua_State* L)
 		}
 		else
 		{
-			if (first && oattr && ((attr | mark) == oattr))
+			if (first && (oattr & (DPY_REVERSE|DPY_UNDERLINE))
+				&& ((attr | mark) == oattr))
 				dpy_writechar(x-1, y, 160); /* non-breaking space */
 
 			dpy_writechar(x, y, c);
@@ -199,21 +200,28 @@ static int prevcharinword_cb(lua_State* L)
 	int offset = luaL_checkint(L, 2) - 1;
 	const char* p = src + offset;
 
-	/* At the beginning of the string? */
+	/* Back up over any control codes; give up if we find ourselves at
+	 * the beginning of the string. */
 
-	if (p == src)
+	do
 	{
-		lua_pushnil(L);
-		return 1;
+		if (p == src)
+		{
+			lua_pushnil(L);
+			return 1;
+		}
+
+		p--;
 	}
+	while (iscntrl(*p));
 
-	/* Back up, and then skip any UTF-8 trailing bytes. */
+	/* We now know that there is a character, and it's definitively not a
+	 * control code. Skip any UTF-8 trailing bytes. */
 
-	p--;
 	while ((p != src) && (getu8bytes(*p) == 0))
 		p--;
 
-	/* Back up over any control codes. */
+	/* Back up over any additional control codes. */
 
 	while (p != src)
 	{
@@ -278,6 +286,7 @@ static int insertintoword_cb(lua_State* L)
 	size_t insbytes;
 	const char* ins = luaL_checklstring(L, 2, &insbytes);
 	int offset = luaL_checkint(L, 3) - 1;
+	int newstate = luaL_checkint(L, 4);
 
 	char dest[srcbytes + insbytes + OVERHEAD];
 	char* p = dest;
@@ -294,8 +303,7 @@ static int insertintoword_cb(lua_State* L)
 
 		if (!copied && ((s - src) >= offset))
 		{
-			int ss = 0;
-			while (copy(&p, &dstate, &ins, &ss, 0, STYLE_ALL))
+			while (copy(&p, &dstate, &ins, &newstate, 0, STYLE_ALL))
 				;
 
 			insend = p - dest;
@@ -363,7 +371,9 @@ static int applystyletoword_cb(lua_State* L)
 	const char* offset2 = src + luaL_checkint(L, 5) - 1;
 	const char* csoffset = src + luaL_checkint(L, 6) - 1;
 
-	char dest[srcbytes];
+	/* Adding a style will add at most two extra bytes to the string
+	 * (on, and then off again). Probably less. */
+	char dest[srcbytes+2];
 	char* p = dest;
 	char* cdoffset = dest;
 
@@ -403,6 +413,39 @@ static int applystyletoword_cb(lua_State* L)
 	return 2;
 }
 
+/* Fetch the style at a particular offset into a word. */
+
+static int getstylefromword_cb(lua_State* L)
+{
+	size_t srcbytes;
+	const char* src = luaL_checklstring(L, 1, &srcbytes);
+	const char* offset = src + luaL_checkint(L, 2) - 1;
+
+	int state = 0;
+	while (src < offset)
+	{
+		int c = readu8(&src);
+		if (c == '\0')
+			break;
+
+		if (iswcntrl(c))
+			state = c & STYLE_ALL;
+	}
+
+	lua_pushnumber(L, state);
+	return 1;
+}
+
+/* Create a raw style byte. */
+
+static int createstylebyte_cb(lua_State* L)
+{
+	int style = luaL_checkint(L, 1);
+	char buffer = style | STYLE_MARKER;
+	lua_pushlstring(L, &buffer, 1);
+	return 1;
+}
+	
 void word_init(void)
 {
 	const static luaL_Reg funcs[] =
@@ -415,6 +458,8 @@ void word_init(void)
 		{ "insertintoword",            insertintoword_cb },
 		{ "deletefromword",            deletefromword_cb },
 		{ "applystyletoword",          applystyletoword_cb },
+		{ "getstylefromword",          getstylefromword_cb },
+		{ "createstylebyte",           createstylebyte_cb },
 		{ NULL,                        NULL }
 	};
 
