@@ -43,6 +43,7 @@ DocumentSetClass =
 	touch = function(self)
 		self.changed = true
 		self.justchanged = true
+		Document:touch()
 	end,
 	
 	clean = function(self)
@@ -135,7 +136,7 @@ DocumentSetClass =
 	setCurrent = function(self, name)
 		-- Ensure any housekeeping on the current document gets done.
 		
-		if Document and Document.changed then
+		if Document.changed then
 			FireEvent(Event.Changed)
 		end
 
@@ -214,14 +215,21 @@ DocumentClass =
 	
 	-- remove any cached data prior to saving
 	purge = function(self)
-		for _, l in ipairs(self) do
-			l:touch()
+		for _, paragraph in ipairs(self) do
+			paragraph:touch()
 		end
+
 		self.topp = nil
 		self.topw = nil
 		self.botp = nil
 		self.botw = nil
 		self.wrapwidth = nil
+
+		-- These should no longer exist; this dates from a previous attempt
+		-- at undo with file version 6. We're not storing the undo buffer
+		-- in files any more.
+		self.undostack = nil
+		self.redostack = nil
 	end,
 	
 	-- calculate space above this paragraph
@@ -259,6 +267,10 @@ DocumentClass =
 			return sb
 		end
 	end,
+
+	touch = function(self)
+		FireEvent(Event.DocumentModified, self)
+	end,
 }
 
 ParagraphClass =
@@ -266,7 +278,7 @@ ParagraphClass =
 	copy = function(self)
 		local words = {}
 		for _, w in ipairs(self) do
-			words[#words+1] = w:copy()
+			words[#words+1] = w
 		end
 		
 		return CreateParagraph(self.style, words)
@@ -275,6 +287,7 @@ ParagraphClass =
 	touch = function(self)
 		self.lines = nil
 		self.wrapwidth = nil
+		self.xs = nil
 	end,
 	
 	wrap = function(self, width)
@@ -285,19 +298,23 @@ ParagraphClass =
 			local lines = {}
 			local line = {wn = 1}
 			local w = 0
+			local xs = {}
+			self.xs = xs
 			
 			for wn, word in ipairs(self) do
-				word.x = w
-				local ww = word:getWidth()
+				-- get width of word (including space)
+				local ww = GetStringWidth(word) + 1
+	
+				xs[wn] = w
 				w = w + ww
 				if (w >= width) then
 					lines[#lines+1] = line
 					line = {wn = wn}
 					w = ww
-					word.x = 0
+					xs[wn] = 0
 				end
 				
-				line[#line+1] = word
+				line[#line+1] = wn
 			end
 			
 			if (#line > 0) then
@@ -312,20 +329,26 @@ ParagraphClass =
 	end,
 
 	renderLine = function(self, line, x, y)
-		width = width or (ScreenWidth - x)
-
 		local cstyle = stylemarkup[self.style.name] or 0
 		local ostyle = 0
-		for wn, w in ipairs(line) do
-			local text = w.text
-			
-			ostyle = WriteStyled(x+w.x, y, text, ostyle, nil, nil, cstyle)
+		local xs = self.xs
+		for _, wn in ipairs(line) do
+			local w = self[wn]
+
+			local payload = {
+				word = w,
+				ostyle = ostyle,
+				cstyle = cstyle
+			}
+			FireEvent(Event.DrawWord, payload)
+
+			ostyle = WriteStyled(x+xs[wn], y, payload.word,
+				payload.ostyle, nil, nil, payload.cstyle)
 		end
 	end,
 
 	renderMarkedLine = function(self, line, x, y, width, pn)
 		width = width or (ScreenWidth - x)
-		marked = marked or false
 		
 		local lwn = line.wn		
 		local mp1, mw1, mo1, mp2, mw2, mo2 = Document:getMarks()
@@ -370,7 +393,15 @@ ParagraphClass =
 				end
 			end
 			
-			ostyle = WriteStyled(x+w.x, y, w.text, ostyle, s, e, cstyle)
+			local payload = {
+				word = self[w],
+				ostyle = ostyle,
+				cstyle = cstyle
+			}
+			FireEvent(Event.DrawWord, payload)
+
+			ostyle = WriteStyled(x+self.xs[w], y, payload.word,
+				payload.ostyle, s, e, payload.cstyle)
 		end
 	end,
 
@@ -397,95 +428,70 @@ ParagraphClass =
 	-- returns: X offset, line number, word number in line
 	getXOffsetOfWord = function(self, wn)
 		local lines = self:wrap()
-		local x = self[wn].x
+		local x = self.xs[wn]
 		local ln, wn = self:getLineOfWord(wn)
 		return x, ln, wn
 	end,
 	
-	deleteWordAt = function(self, pos)
-		table.remove(self, pos)
-		self:touch()
-	end,
-	
-	insertWordBefore = function(self, pos, word)
-		table.insert(self, pos, word)
-		self:touch()
-	end,
-	
-	appendWord = function(self, word)
-		self[#self+1] = word
-		self:touch()
-	end,
-	
-	appendWords = function(self, words)
-		for _, w in ipairs(words) do
-			self:appendWord(w)
+	sub = function(self, start, count)
+		if not count then
+			count = #self - start + 1
+		else
+			count = min(count, #self - start + 1)
 		end
-		self:touch()
-	end,
-	
-	split = function(self, wn)
-		local p1 = CreateParagraph(self.style)
-		local p2 = CreateParagraph(self.style)
-		
-		for i=1, wn-1 do
-			p1:appendWord(self[i])
+
+		local t = {}
+		for i = start, start+count-1 do
+			t[#t+1] = self[i]
 		end
-		
-		for i=wn, #self do
-			p2:appendWord(self[i])
-		end
-		
-		return p1, p2
+		return t
 	end,
-	
-	truncateAtWord = function(self, wn)
-		while (self[wn]) do
-			table.remove(self, wn)
-		end
-		self:touch()
-	end,
-	
-	changeStyle = function(self, style)
-		self.style = style
-		self:touch()
-	end,
-	
+
 	-- return an unstyled string containing the contents of the paragraph.
 	asString = function(self)
 		local s = {}
 		for _, w in ipairs(self) do
-			s[#s+1] = w:asString()
+			s[#s+1] = GetWordText(w)
 		end
 		
 		return table_concat(s, " ")
 	end
 }
 
-WordClass =
-{
-	copy = function(self)
-		return CreateWord(self.text)
-	end,
-			
-	getWidth = function(self)
-		return GetStringWidth(self.text) + 1
-	end,
-	
-	-- converts byte offset to X position
-	getXOffsetOfChar = function(self, o)
-		return GetStringWidth(self.text:sub(1, o-1))
-	end,
-	
-	-- converts X position to byte offset
-	getByteOfChar = function(self, x)
-		local text = self.text
-		
-		local len = text:len()
+function CreateParagraph(style, ...)
+	words = {}
+
+	for _, t in ipairs({...}) do
+		if (type(t) == "table") then
+			for _, w in ipairs(t) do
+				words[#words+1] = w
+			end
+		else
+			words[#words+1] = t
+		end
+	end
+
+	words.style = style or DocumentSet.styles["P"]
+	setmetatable(words, {__index = ParagraphClass})
+	return words
+end
+
+-- Returns how many screen spaces a portion of a string takes up.
+function GetWidthFromOffset(s, o)
+	return GetStringWidth(s:sub(1, o-1))
+end
+
+-- Returns the offset into a string needed for a screen width.
+function GetOffsetFromWidth(s, x)
+		local len = #s
 		local o = 1
 		while (o <= len) do
-			local charlen = GetBytesOfCharacter(string.byte(text, o))
-			local char = text:sub(o, o+charlen-1)
+			if (x == 0) then
+				return o
+			end
+
+			local charlen = GetBytesOfCharacter(string.byte(s, o))
+			local char = s:sub(o, o+charlen-1)
 			local ww = GetStringWidth(char)
 			if (ww > x) then
 				return o
@@ -496,13 +502,17 @@ WordClass =
 		end
 		
 		return len + 1
-	end,
-	
-	-- returns an unstyled string containing the word contents
-	asString = function(self)
-		return GetWordText(self.text)
-	end,
-}
+end
+
+function GetWordSimpleText(s)
+	s = GetWordText(s)
+	s = UnSmartquotify(s)
+	s = s:gsub('[~#&^$"<>]+', "")
+	s = s:gsub("^[.'([{]+", "")
+	s = s:gsub("[',.!?:;)%]}]+$", "")
+	s = s:lower()
+	return s
+end
 
 local function create_styles()
 	local styles =
@@ -600,6 +610,10 @@ local function create_styles()
 	return styles
 end
 	
+function ResetParagraphStyles()
+	DocumentSet.styles = create_styles()
+end
+
 function CreateDocumentSet()
 	local ds =
 	{
@@ -628,24 +642,8 @@ function CreateDocument()
 	
 	setmetatable(d, {__index = DocumentClass})
 	
-	local p = CreateParagraph(DocumentSet.styles["P"], {CreateWord()})
+	local p = CreateParagraph(DocumentSet.styles["P"], {""})
 	d:appendParagraph(p)
 	return d
 end
 
-function CreateParagraph(style, words)
-	words = words or {}
-	words.style = style or DocumentSet.styles["P"]
-	setmetatable(words, {__index = ParagraphClass})
-	return words
-end
-
-function CreateWord(text)
-	local w =
-	{
-		text = text or ""
-	}
-	
-	setmetatable(w, {__index = WordClass})
-	return w
-end
