@@ -16,18 +16,41 @@ local redrawpending = true
 
 HOME = os.getenv("HOME") or os.getenv("USERPROFILE")
 
+-- Determine the installation directory (Windows only).
+
+if (ARCH == "windows") then
+	local exe = os.getenv("WINDOWS_EXE")
+	local _, _, dir = exe:find("^(.*)[/\\][^/\\]*$")
+	WINDOWS_INSTALL_DIR = dir
+end
+
+-- Which config file are we loading?
+
+local configfile = HOME .. "/.wordgrinder.lua"
+
+local oldcp, oldcw, oldco
 function QueueRedraw()
 	redrawpending = true
-	if not Document.wrapwidth then
-		ResizeScreen()
+	if Document then
+		if (oldcp ~= Document.cp) or (oldcw ~= Document.cw) or
+				(oldco ~= Document.co) then
+			oldcp = Document.cp
+			oldcw = Document.cw
+			oldco = Document.co
+			FireEvent(Event.Moved)
+		end
+
+		if not Document.wrapwidth then
+			ResizeScreen()
+		end
 	end
 end
 
 function ResetDocumentSet()
 	DocumentSet = CreateDocumentSet()
 	DocumentSet.menu = CreateMenu()
+	Document = CreateDocument()
 	DocumentSet:addDocument(CreateDocument(), "main")
-	DocumentSet:setCurrent("main")
 	RebuildParagraphStylesMenu(DocumentSet.styles)
 	RebuildDocumentsMenu(DocumentSet.documents)
 	DocumentSet:purge()
@@ -37,16 +60,55 @@ function ResetDocumentSet()
 	FireEvent(Event.RegisterAddons)
 end
 
+do
+	local function cb(event, token)
+		oldcp = 1
+		oldcw = 1
+		oldco = 1
+		SetCurrentStyleHint(0, 0)
+	end
+
+	AddEventListener(Event.DocumentLoaded, cb)
+	AddEventListener(Event.DocumentCreated, cb)
+end
+	
+-- Kick the garbage collector whenever we're idle, just to keep
+-- memory usage down.
+
+do
+	local function cb(event)
+		collectgarbage("collect")
+		QueueRedraw()
+	end
+
+	AddEventListener(Event.Idle, cb)
+end
+
 -- This function contains the word processor proper, including the main event
 -- loop.
 
 function WordProcessor(filename)
+	LoadGlobalSettings()
+
+	do
+		local f, e = loadfile(configfile)
+		if e and not e:find("No such file or directory") then
+			CLIError("config file compilation error: "..e)
+		end
+
+		if not e then
+			xpcall(f, CLIError)
+		end
+	end
+
 	wg.initscreen()
 	ResizeScreen()
 	RedrawScreen()
 	
 	if filename then
 		Cmd.LoadDocumentSet(filename)
+	else
+		FireEvent(Event.DocumentLoaded)
 	end
 	
 	--ModalMessage("Welcome!", "Welcome to WordGrinder! While editing, you may press ESC for the menu, or ESC, F, X to exit (or ALT+F, X if your terminal supports it).")
@@ -59,8 +121,10 @@ function WordProcessor(filename)
 		
 		["KEY_REDRAW"] = RedrawScreen,
 		
-		[" "] = Cmd.SplitCurrentWord,
-		["KEY_RETURN"] = Cmd.SplitCurrentParagraph,
+		[" "] = { Cmd.Checkpoint, Cmd.TypeWhileSelected,
+			Cmd.SplitCurrentWord },
+		["KEY_RETURN"] = { Cmd.Checkpoint, Cmd.TypeWhileSelected,
+			Cmd.SplitCurrentParagraph },
 		["KEY_ESCAPE"] = Cmd.ActivateMenu,
 	}	
 		
@@ -72,6 +136,7 @@ function WordProcessor(filename)
 				DocumentSet.justchanged = false
 			end
 			
+			FlushAsyncEvents()
 			FireEvent(Event.WaitingForUser)
 			local c = "KEY_TIMEOUT"
 			while (c == "KEY_TIMEOUT") do
@@ -91,21 +156,23 @@ function WordProcessor(filename)
 			-- Anything in masterkeymap overrides everything else.
 			local f = masterkeymap[c]
 			if f then
-				f()
+				RunMenuAction(f)
 			else
 				-- It's not in masterkeymap. If it's printable, insert it; if it's
 				-- not, look it up in the menu hierarchy.
 				
 				if not c:match("^KEY_") then
-					Cmd.InsertStringIntoWord(c)
+					Cmd.Checkpoint()
+					Cmd.TypeWhileSelected()
+
+					local payload = { value = c }
+					FireEvent(Event.KeyTyped, payload)
+
+					Cmd.InsertStringIntoWord(payload.value)
 				else
 					f = DocumentSet.menu:lookupAccelerator(c)
 					if f then
-						if (type(f) == "function") then
-							f()
-						else
-							Cmd.ActivateMenu(f)
-						end
+						RunMenuAction(f)
 					else
 						NonmodalMessage(c:gsub("^KEY_", "").." is not bound --- try ESCAPE for a menu")
 					end
@@ -117,6 +184,7 @@ function WordProcessor(filename)
 	while true do
 		local f, e = xpcall(eventloop, Traceback)
 		if not f then
+			print(e)
 			ModalMessage("Internal error!",
 				"Something went wrong inside WordGrinder! I'll try and "..
 				"continue but you should save your work immediately (under a "..
@@ -139,16 +207,6 @@ function Main(...)
 		local stdout = io.stdout
 		local stderr = io.stderr
 		
-		local function message(...)
-			stderr:write("wordgrinder: ", ...)
-			stderr:write("\n")
-		end
-	
-		local function usererror(...)
-			message(...)
-			os.exit(1)
-		end
-		
 		local function do_help(opt)
 			stdout:write("WordGrinder version ", VERSION, " © 2007-2008 David Given\n")
 			if DEBUG then
@@ -158,9 +216,11 @@ function Main(...)
 			stdout:write([[
 Syntax: wordgrinder [<options...>] [<filename>]
 Options:
-   -h    --help              Displays this message.
-         --lua file.lua      Loads and executes file.lua before startup
-   -c    --convert src dest  Converts from one file format to another
+   -h    --help                Displays this message.
+         --lua file.lua        Loads and executes file.lua and then exits
+		                       (or file.lua can be some Lua statements)
+   -c    --convert src dest    Converts from one file format to another
+         --config file.lua     Sets the name of the user config file
 
 Only one filename may be specified, which is the name of a WordGrinder
 file to load on startup. If not given, you get a blank document instead.
@@ -169,6 +229,13 @@ To convert documents, use --convert. The file type is autodetected from the
 extension. To specify a document name, use :name as a suffix. e.g.:
 
     wordgrinder --convert filename.wg:"Chapter 1" chapter1.odt
+
+The user config file is a Lua file which is loaded and executed before
+the program starts up (but after any --lua files). It defaults to:
+
+    ]] .. configfile .. [[
+
+
 ]])
 			if DEBUG then
 				-- List debugging options here.
@@ -179,27 +246,41 @@ extension. To specify a document name, use :name as a suffix. e.g.:
 		
 		local function do_lua(opt1, opt2)
 			if not opt1 then
-				usererror("--lua must have an argument")
+				CLIError("--lua must have an argument")
 			end
-			
+
 			local f, e = loadfile(opt1)
-			if e then
-				usererror("user script compilation error: "..e)
+			if e and e:find("No such file or directory") then
+				f, e = loadstring(opt1)
 			end
-			return f(opt2) or 1
+			if e then
+				CLIError("user script compilation error: "..e)
+			end
+
+			f(opt2)
+			os.exit(0)
 		end
 		
 		local function do_convert(opt1, opt2)
 			if not opt1 or not opt2 then
-				usererror("--convert must have two arguments")
+				CLIError("--convert must have two arguments")
 			end
 			
 			CliConvert(opt1, opt2)
 		end
 		
+		local function do_config(opt1, opt2)
+			if not opt1 then
+				CLIError("--config must have an argument")
+			end
+
+			configfile = opt1
+			return 1
+		end
+
 		local function needarg(opt)
 			if not opt then
-				usererror("missing option parameter")
+				CLIError("missing option parameter")
 			end
 		end
 		
@@ -209,6 +290,7 @@ extension. To specify a document name, use :name as a suffix. e.g.:
 			["lua"]         = do_lua,
 			["c"]           = do_convert,
 			["convert"]     = do_convert,
+			["config"]      = do_config,
 		}
 		
 		if DEBUG then
@@ -219,7 +301,7 @@ extension. To specify a document name, use :name as a suffix. e.g.:
 		-- Called on an unrecognised option.
 		
 		local function unrecognisedarg(arg)
-			usererror("unrecognised option '", arg, "' --- try --help for help")
+			CLIError("unrecognised option '", arg, "' --- try --help for help")
 		end
 		
 		-- Do the actual argument parsing.
@@ -256,7 +338,7 @@ extension. To specify a document name, use :name as a suffix. e.g.:
 				end
 			else
 				if filename then
-					usererror("you may only specify one filename")
+					CLIError("you may only specify one filename")
 				end
 				filename = o
 			end	
