@@ -10,9 +10,13 @@ local PrevCharInWord = wg.prevcharinword
 local InsertIntoWord = wg.insertintoword
 local DeleteFromWord = wg.deletefromword
 local ApplyStyleToWord = wg.applystyletoword
+local GetStyleFromWord = wg.getstylefromword
+local CreateStyleByte = wg.createstylebyte
 local ReadU8 = wg.readu8
 local WriteU8 = wg.writeu8
+local P = M.P
 local table_concat = table.concat
+local unpack = unpack or table.unpack
 
 function Cmd.GotoBeginningOfWord()
 	Document.co = 1
@@ -21,7 +25,7 @@ function Cmd.GotoBeginningOfWord()
 end
 
 function Cmd.GotoEndOfWord()
-	Document.co = Document[Document.cp][Document.cw].text:len() + 1
+	Document.co = #Document[Document.cp][Document.cw] + 1
 	QueueRedraw()
 	return true
 end
@@ -125,7 +129,7 @@ end
 
 function Cmd.GotoPreviousChar()
 	local word = Document[Document.cp][Document.cw]
-	local co = PrevCharInWord(word.text, Document.co)
+	local co = PrevCharInWord(word, Document.co)
 	if not co then
 		return false
 	end
@@ -138,7 +142,7 @@ end
 
 function Cmd.GotoNextChar()
 	local word = Document[Document.cp][Document.cw]
-	local co = NextCharInWord(word.text, Document.co)
+	local co = NextCharInWord(word, Document.co)
 	if not co then
 		return false
 	end
@@ -164,17 +168,20 @@ function Cmd.GotoNextCharW()
 end
 
 function Cmd.InsertStringIntoWord(c)
-	local paragraph = Document[Document.cp]
-	local word = paragraph[Document.cw]
+	local cp, cw, co = Document.cp, Document.cw, Document.co
+	local paragraph = Document[cp]
+	local word = paragraph[cw]
 	
-	local s, co = InsertIntoWord(word.text, c, Document.co)
+	local s, co = InsertIntoWord(word, c, co, GetCurrentStyleHint())
 	if not co then
 		return false
 	end
 	
-	word.text = s
+	Document[cp] = CreateParagraph(paragraph.style,
+		paragraph:sub(1, cw-1),
+		s,
+		paragraph:sub(cw+1))
 	Document.co = co
-	paragraph:touch()
 	
 	DocumentSet:touch()
 	QueueRedraw()
@@ -196,19 +203,48 @@ function Cmd.InsertStringIntoParagraph(c)
 end
 
 function Cmd.SplitCurrentWord()
-	local paragraph = Document[Document.cp]
-	local word = paragraph[Document.cw]
-	local text = word.text
-	local co = Document.co
-	
-	local left = DeleteFromWord(text, co, text:len()+1)
-	local newword = CreateWord(left)
+	local cp, cw, co = Document.cp, Document.cw, Document.co
+	local styleprime = ""
+	local styleprimelen = 0
 
-	word.text = DeleteFromWord(text, 1, co)
-	
-	paragraph:insertWordBefore(Document.cw, newword)
-	Document.cw = Document.cw + 1
-	Document.co = 1
+	if not Document.mp then
+		local stylehint = GetCurrentStyleHint()
+		-- This is a bit evil. We want to prime the new word with the current
+		-- style hint, so that when the cursor moves we don't lose the user's
+		-- style hint in favour of the one read from the old word, which will
+		-- be stale.
+		--
+		-- However, we don't want to insert any actual *text* in the new style.
+		-- This means that the frameworks we have for applying style won't
+		-- work. Instead, we exploit our knowledge of how the style bytes
+		-- are implemented to do it manually.
+		-- 
+		-- We *also* don't want to change the actual style of the text. So the
+		-- prime code needs to be followed by an unprime code. The cursor will
+		-- be placed between these. (As soon as the user types, all this
+		-- insanity will be undone.)
+		
+		styleprime = CreateStyleByte(stylehint)
+		if (stylehint ~= 0) then
+			-- Only add this is needed to prevent stupid control code buildup.
+			styleprime = styleprime .. CreateStyleByte(0)
+		end
+		styleprimelen = 1
+	end
+
+	local paragraph = Document[cp]
+	local word = paragraph[cw]
+	local left = DeleteFromWord(word, co, #word+1)
+	local right = DeleteFromWord(word, 1, co)
+
+	Document[cp] = CreateParagraph(paragraph.style,
+		paragraph:sub(1, cw-1),
+		left,
+		styleprime..right,
+		paragraph:sub(cw+1))
+
+	Document.cw = cw + 1
+	Document.co = 1 + styleprimelen
 	
 	DocumentSet:touch()
 	QueueRedraw()
@@ -216,12 +252,15 @@ function Cmd.SplitCurrentWord()
 end
 
 function Cmd.JoinWithNextParagraph()
-	if (Document.cp == #Document) then
+	local cp, cw, co = Document.cp, Document.cw, Document.co
+	if (cp == #Document) then
 		return false
 	end
 	
-	Document[Document.cp]:appendWords(Document[Document.cp+1])
-	Document:deleteParagraphAt(Document.cp+1)
+	Document[cp] = CreateParagraph(Document[cp].style,
+		Document[cp],
+		Document[cp+1])
+	Document:deleteParagraphAt(cp+1)
 	
 	DocumentSet:touch()
 	QueueRedraw()
@@ -229,19 +268,21 @@ function Cmd.JoinWithNextParagraph()
 end
 
 function Cmd.JoinWithNextWord()
-	local paragraph = Document[Document.cp]
-	local word = paragraph[Document.cw]
+	local cp, cw, co = Document.cp, Document.cw, Document.co
+	local paragraph = Document[cp]
 
-	if (Document.cw == #paragraph) then
+	if (cw == #paragraph) then
 		if not Cmd.JoinWithNextParagraph() then
 			return false
 		end
+		paragraph = Document[cp]
 	end
 
-	if (Document.cw ~= #paragraph) then
-		word.text = InsertIntoWord(paragraph[Document.cw+1].text, word.text, 1)
-		paragraph:deleteWordAt(Document.cw+1)
-	end
+	local word = InsertIntoWord(paragraph[cw+1], paragraph[cw], 1, 0)
+	Document[cp] = CreateParagraph(paragraph.style,
+		paragraph:sub(1, cw-1),
+		word,
+		paragraph:sub(cw+2))
 	
 	DocumentSet:touch()
 	QueueRedraw()
@@ -252,67 +293,128 @@ function Cmd.DeletePreviousChar()
 	return Cmd.GotoPreviousCharW() and Cmd.DeleteNextChar()
 end
 
+function Cmd.DeleteSelectionOrPreviousChar()
+	if Document.mp then
+		return Cmd.Delete()
+	else
+		return Cmd.DeletePreviousChar()
+	end
+end
+
 function Cmd.DeleteNextChar()
-	local paragraph = Document[Document.cp]
-	local word = paragraph[Document.cw]
-	local co = Document.co
-	local nextco = NextCharInWord(word.text, co)
+	local cp, cw, co = Document.cp, Document.cw, Document.co
+	local paragraph = Document[cp]
+	local word = paragraph[cw]
+	local nextco = NextCharInWord(word, co)
 	if not nextco then
 		return Cmd.JoinWithNextWord()
 	end
 	
-	word.text = DeleteFromWord(word.text, co, nextco)
-	paragraph:touch()
+	Document[cp] = CreateParagraph(paragraph.style,
+		paragraph:sub(1, cw-1),
+		DeleteFromWord(word, co, nextco),
+		paragraph:sub(cw+1))
 	
 	DocumentSet:touch()
 	QueueRedraw()
 	return true
 end
 
+function Cmd.DeleteSelectionOrNextChar()
+	if Document.mp then
+		return Cmd.Delete()
+	else
+		return Cmd.DeleteNextChar()
+	end
+end
+
+function Cmd.DeleteWordLeftOfCursor()
+	local cp = Document.cp
+	local cw = Document.cw
+	local co = Document.co
+	local paragraph = Document[cp]
+	local word = paragraph[cw]
+
+	Document[cp] = CreateParagraph(paragraph.style,
+		paragraph:sub(1, cw-1),
+		DeleteFromWord(word, 1, co),
+		paragraph:sub(cw+1))
+	Document.co = 1
+	
+	DocumentSet:touch()
+	QueueRedraw()
+	return true
+end
+
+function Cmd.DeleteWord()
+	if (Document.co == 1) and (Document.cw == 1) then
+		return Cmd.DeletePreviousChar()
+	end
+
+	if (Document.co == 1) then
+		Cmd.DeletePreviousChar()
+	end
+	return Cmd.DeleteWordLeftOfCursor()
+end
+
 function Cmd.SplitCurrentParagraph()
-	Cmd.SplitCurrentWord()
+	if (Document.co == 1) and (Document.cw == 1) then
+		-- Beginning of paragraph; we're going to create a new, empty paragraph
+		-- so we need to split to make sure there's an empty word for it to
+		-- contain.
+		Cmd.SplitCurrentWord()
+	elseif (Document.co > 1) then
+		-- Otherwise, only split if we're not at the beginning of a word.
+		Cmd.SplitCurrentWord()
+	else
+		DocumentSet:touch()
+		QueueRedraw()
+	end
 	
-	local p1, p2 = Document[Document.cp]:split(Document.cw)
+	local cp, cw = Document.cp, Document.cw
+	local paragraph = Document[cp]
+	local p1 = CreateParagraph(paragraph.style, paragraph:sub(1, cw-1))
+	local p2 = CreateParagraph(paragraph.style, paragraph:sub(cw))
 	
-	Document:deleteParagraphAt(Document.cp)
-	Document:insertParagraphBefore(p2, Document.cp)
-	Document:insertParagraphBefore(p1, Document.cp)
+	Document[cp] = p2
+	Document:insertParagraphBefore(p1, cp)
 	Document.cp = Document.cp + 1
 	Document.cw = 1
 	Document.co = 1
+	QueueRedraw()
 	return true
 end
 
 function Cmd.GotoXPosition(pos)
 	local paragraph = Document[Document.cp]
-	local lines = paragraph:wrap()
+	local lines = paragraph:wrap(Document.wrapwidth)
 	local ln = paragraph:getLineOfWord(Document.cw)
 	
-	local l = lines[ln]
-	local wn = #l
-	local word
+	local line = lines[ln]
+	local wordofline = #line
 
 	pos = pos - (paragraph.style.indent or 0)
 	if (pos < 0) then
 		pos = 0
 	end
 	
-	while (wn > 0) do
-		word = l[wn]
-		
-		if (word.x <= pos) then
+	while (wordofline > 0) do
+		if (paragraph.xs[line[wordofline]] <= pos) then
 			break
 		end
-		wn = wn - 1
+		wordofline = wordofline - 1
 	end
 	
-	if (wn == 0) then
-		wn = 1
+	if (wordofline == 0) then
+		wordofline = 1
 	end
 
-	wo = word:getByteOfChar(pos - word.x) 
+	local wn = line[wordofline]
+	local word = paragraph[wn]
+	local wordx = paragraph.xs[wn]
+	wo = GetOffsetFromWidth(word, pos - wordx)
 	
-	Document.cw = paragraph:getWordOfLine(ln) + wn - 1
+	Document.cw = paragraph:getWordOfLine(ln) + wordofline - 1
 	Document.co = wo
 
 	QueueRedraw()
@@ -325,7 +427,7 @@ local function getpos()
 	local cw = Document.cw
 	local word = paragraph[cw]
 	local x, ln, wn = paragraph:getXOffsetOfWord(cw)
-	x = x + word:getXOffsetOfChar(Document.co)
+	x = x + GetWidthFromOffset(word, Document.co)
 	x = x + (paragraph.style.indent or 0)
 
 	return x, ln, lines
@@ -335,6 +437,10 @@ function Cmd.GotoNextLine()
 	local x, ln, lines = getpos()
 
 	if (ln == #lines) then
+		if (Document.cp == #Document) then
+			return Cmd.GotoEndOfParagraph()
+		end
+
 		return Cmd.GotoNextParagraph() and
 		       Cmd.GotoBeginningOfParagraph() and
 		       Cmd.GotoXPosition(x)
@@ -348,6 +454,10 @@ function Cmd.GotoPreviousLine()
 	local x, ln, lines = getpos()
 
 	if (ln == 1) then
+		if (Document.cp == 1) then
+			return Cmd.GotoBeginningOfParagraph()
+		end
+
 		return Cmd.GotoPreviousParagraph() and
 		       Cmd.GotoEndOfParagraph() and
 		       Cmd.GotoXPosition(x)
@@ -395,7 +505,7 @@ local style_tab =
 	["o"] = {0, 0},
 }
 
-function Cmd.ToggleStyle(s)
+function Cmd.ApplyStyleToSelection(s)
 	if not Document.mp then
 		return false
 	end
@@ -416,28 +526,34 @@ function Cmd.ToggleStyle(s)
 			lastword = mw2
 		end
 		
-		for w = firstword, lastword do
-			local word = paragraph[w]
+		local words = {}
+		for wn = firstword, lastword do
+			local word = paragraph[wn]
 			
 			local fo = 1
-			local lo = word.text:len() + 1
+			local lo = #word + 1
 			
-			if (p == mp1) and (w == mw1) then
+			if (p == mp1) and (wn == mw1) then
 				fo = mo1
 			end
 			
-			if (p == mp2) and (w == mw2) then
+			if (p == mp2) and (wn == mw2) then
 				lo = mo2
 			end
 			
-			if (p == cp) and (w == cw) then
-				word.text, Document.co = ApplyStyleToWord(word.text, sor, sand, fo, lo, co)
+			if (p == cp) and (wn == cw) then
+				word, Document.co = ApplyStyleToWord(word, sor, sand, fo, lo, co)
 			else
-				word.text = ApplyStyleToWord(word.text, sor, sand, fo, lo, 0)
+				word = ApplyStyleToWord(word, sor, sand, fo, lo, 0)
 			end
+
+			words[#words+1] = word
 		end
 		
-		paragraph:touch()
+		Document[p] = CreateParagraph(paragraph.style,
+			paragraph:sub(1, firstword-1),
+			words,
+			paragraph:sub(lastword+1))
 	end
 	
 	Cmd.UnsetMark()
@@ -446,6 +562,41 @@ function Cmd.ToggleStyle(s)
 	return true
 end
 
+function Cmd.SetStyle(s)
+	if Document.mp then
+		return Cmd.ApplyStyleToSelection(s)
+	end
+
+	local sor, sand = unpack(style_tab[s])
+	SetCurrentStyleHint(sor, sand)
+	QueueRedraw()
+	return true;
+end
+
+function GetStyleAtCursor()
+	local cp = Document.cp
+	local cw = Document.cw
+	local co = Document.co
+
+	return GetStyleFromWord(Document[cp][cw], co)
+end
+
+function GetStyleToLeftOfCursor()
+	local cp = Document.cp
+	local cw = Document.cw
+	local co = Document.co
+
+	if (co == 1) then
+		if (cw == 1) then
+			return 0
+		end
+		cw = cw - 1
+		co = #Document[cp][cw]
+	end
+
+	return GetStyleFromWord(Document[cp][cw], co)
+end
+
 function Cmd.ActivateMenu(menu)
 	DocumentSet.menu:activate(menu)
 	ResizeScreen()
@@ -495,30 +646,23 @@ local function styles()
 end
 	
 function Cmd.ChangeParagraphStyle(style)
-	local paragraph = Document[Document.cp]
-	if not style then
-		-- style = PromptForString("Change paragraph style", "Please enter the new paragraph style:", paragraph.style.name)
-		style = Browser("Change paragraph style", "Please select the new paragraph style from the list, or enter a style name:", "Style:", styles())
-		
-	end
-	if not style then
-		return false
-	end
-	
 	style = DocumentSet.styles[style]
 	if not style then
 		ModalMessage("Unknown paragraph style", "Sorry! I don't recognise that style. (This user interface will be improved.)")
 		return false
 	end
 
+	local first, last
 	if Document.mp then	
-		local mp1, _, _, mp2, _, _ = Document:getMarks()
-		
-		for p = mp1, mp2 do
-			Document[p]:changeStyle(style)
-		end
+		local _
+		first, _, _, last, _, _ = Document:getMarks()
 	else
-		paragraph:changeStyle(style)
+		first = Document.cp
+		last = first
+	end
+		
+	for p = first, last do
+		Document[p] = CreateParagraph(style, Document[p])
 	end
 	
 	DocumentSet:touch()
@@ -531,16 +675,28 @@ function Cmd.ToggleMark()
 		Document.mp = nil
 		Document.mw = nil
 		Document.mo = nil
+		Document.sticky_selection = false
 	else
 		Document.mp = Document.cp
 		Document.mw = Document.cw
 		Document.mo = Document.co
+		Document.sticky_selection = true
 	end
 	
 	QueueRedraw()
 	return true
 end
 
+function Cmd.SetMark()
+	if not Document.mp then
+		Document.mp = Document.cp
+		Document.mw = Document.cw
+		Document.mo = Document.co
+		Document.sticky_selection = false
+	end
+	return true
+end
+
 function Cmd.UnsetMark()
 	Document.mp = nil
 	Document.mw = nil
@@ -550,6 +706,27 @@ function Cmd.UnsetMark()
 	return true
 end
 
+function Cmd.MoveWhileSelected()
+	if Document.mp and not Document.sticky_selection then
+		return Cmd.UnsetMark()
+	end
+	return true
+end
+
+function Cmd.TypeWhileSelected()
+	if Document.mp and not Document.sticky_selection then
+		return Cmd.Delete()
+	end
+	return true
+end
+
+function Cmd.SelectWord()
+	return Cmd.UnsetMark() and
+		Cmd.GotoBeginningOfWord() and
+		Cmd.SetMark() and
+		Cmd.GotoEndOfWord()
+end
+
 function Cmd.ChangeDocument(name)
 	if not DocumentSet:findDocument(name) then
 		return false
@@ -587,31 +764,23 @@ function Cmd.Copy(keepselection)
 	
 	-- Remove any words in the first paragraph that weren't copied.
 	
-	local paragraph = buffer[1]
-	while (mw1 > 1) do
-		paragraph:deleteWordAt(1)
-		mw1 = mw1 - 1
+	local paragraph
+	if (mw1 > 1) then
+		paragraph = buffer[1]
+		buffer[1] = CreateParagraph(paragraph.style,
+			paragraph:sub(mw1))
 		if (mp1 == mp2) then
-			mw2 = mw2 - 1
+			mw2 = mw2 - mw1 + 1
 		end
+		mw1 = 1
 	end
 	
 	-- Remove any words in the last paragraph that weren't copied.
 	
 	paragraph = buffer[#buffer]
-	while (mw2 < #paragraph) do
-		paragraph:deleteWordAt(#paragraph)
-	end
-	
-	-- Remove any characters in the leading word that weren't copied.
-	
-	paragraph = buffer[1]
-	word = paragraph[1]
-	if word then
-		word.text = DeleteFromWord(word.text, 1, mo1)
-		if (mp1 == mp2) and (mw1 == mw2) then
-			mo2 = mo2 - mo1 + 1
-		end
+	if (mw2 < #paragraph) then
+		buffer[#buffer] = CreateParagraph(paragraph.style,
+			paragraph:sub(1, mw2))
 	end
 	
 	-- Remove any characters in the trailing word that weren't copied.
@@ -619,7 +788,19 @@ function Cmd.Copy(keepselection)
 	paragraph = buffer[#buffer]
 	word = paragraph[#paragraph]
 	if word then
-		word.text = DeleteFromWord(word.text, mo2, word.text:len()+1)
+		buffer[#buffer] = CreateParagraph(paragraph.style,
+			paragraph:sub(1, #paragraph-1),
+			DeleteFromWord(word, mo2, word:len()+1))
+	end
+	
+	-- Remove any characters in the leading word that weren't copied.
+	
+	paragraph = buffer[1]
+	local word = paragraph[1]
+	if word then
+		buffer[1] = CreateParagraph(paragraph.style,
+			{DeleteFromWord(word, 1, mo1)},
+			paragraph:sub(2))
 	end
 	
 	NonmodalMessage("Selected area copied to clipboard.")
@@ -646,10 +827,13 @@ function Cmd.Paste()
 	local cw = Document.cw
 	Cmd.SplitCurrentWord()
 	local paragraph = Document[Document.cp]
-	for _, word in ipairs(buffer[1]) do
-		paragraph:insertWordBefore(Document.cw, word:copy())
-		Document.cw = Document.cw + 1
-	end
+
+	Document[Document.cp] = CreateParagraph(paragraph.style,
+		paragraph:sub(1, cw),
+		buffer[1],
+		paragraph:sub(cw+1))
+	Document.cw = Document.cw + #buffer[1]
+	Document.co = 1
 	
 	-- Splice the first word of the section just pasted.
 	
@@ -669,18 +853,22 @@ function Cmd.Paste()
 		
 		local p = 2
 		for p = 2, #buffer do	
-			local copy = buffer[p]:copy()
-			Document:insertParagraphBefore(copy, Document.cp)
+			local paragraph = buffer[p]
+			Document:insertParagraphBefore(
+				CreateParagraph(paragraph.style, paragraph),
+				Document.cp)
+
 			Document.cp = Document.cp + 1
 			Document.cw = 1
 			Document.co = 1
 		end
 	end
-	
+
 	-- Splice the last word of the section just pasted.
 	
 	NonmodalMessage("Clipboard copied to cursor position.")
-	return Cmd.GotoPreviousCharW() and Cmd.JoinWithNextWord()
+	return Cmd.GotoBeginningOfWord() and Cmd.GotoPreviousCharW()
+		and Cmd.JoinWithNextWord()
 end
 
 function Cmd.Delete()
@@ -717,10 +905,22 @@ function Cmd.Delete()
 	
 	-- And merge the two areas together again.
 	
+	if not (Cmd.GotoPreviousWordW() and
+	        Cmd.GotoEndOfWord() and
+			Cmd.JoinWithNextWord()) then
+		return false
+	end
+
+	-- If the selection started at a word boundary, make sure it's preserved.
+	
+	if (mo1 == 1) and (mw1 > 1) then
+		if not Cmd.SplitCurrentWord() then
+			return false
+		end
+	end
+	
 	NonmodalMessage("Selected area deleted.")
-	return Cmd.GotoPreviousCharW() and
-	       Cmd.JoinWithNextWord() and
-	       Cmd.UnsetMark()
+	return Cmd.UnsetMark()
 end
 
 function Cmd.Find(findtext, replacetext)
@@ -731,10 +931,17 @@ function Cmd.Find(findtext, replacetext)
 		end
 	end
 
-	-- Convert the search text into a pattern.
-	
-	local patterns = {}
-	local words = SplitString(findtext, "%s")
+	DocumentSet.findtext = findtext
+	DocumentSet._findpatterns = nil
+	DocumentSet.replacetext = replacetext
+	return Cmd.FindNext()	
+end
+
+local function compile_patterns(text)
+	patterns = {}
+	local words = SplitString(text, "%s")
+	local smartquotes = DocumentSet.addons.smartquotes or {}
+
 	for _, w in ipairs(words) do
 		-- w is a word from the pattern. We need to perform the following
 		-- changes:
@@ -755,27 +962,37 @@ function Cmd.Find(findtext, replacetext)
 
 			if ((c >= "A") and (c <= "Z")) or
 			    ((c >= "a") and (c <= "z")) then
-				c = "["..c:upper()..c:lower().."]"
+				c = P("["..c:upper()..c:lower().."]")
+			elseif (c == "'") then
+				c = P("'") + P(smartquotes.leftsingle) + P(smartquotes.rightsingle)
+			elseif (c == '"') then
+				c = P('"') + P(smartquotes.leftdouble) + P(smartquotes.rightdouble)
 			end
 			
-			wp[#wp+1] = c
+			wp[#wp+1] = P(c)
 		end
 		
-		patterns[#patterns + 1] = table_concat(wp, "%c*")
+		patterns[#patterns + 1] = P(unpack(Intersperse(wp, "%c*")))
 	end
 	
 	for i = 2, (#patterns - 1) do
-		patterns[i] = "^"..patterns[i].."$"
+		patterns[i] = P("^%c*", patterns[i], "%c*$")
 	end
 	
 	if (#patterns > 1) then
-		patterns[1] = patterns[1].."$"
-		patterns[#patterns] = "^"..patterns[#patterns]
+		patterns[1] = P(patterns[1], "%c*$")
+		patterns[#patterns] = P("^%c*", patterns[#patterns])
 	end
-	
-	DocumentSet.findtext = patterns
-	DocumentSet.replacetext = replacetext
-	return Cmd.FindNext()	
+
+	for i = 1, #patterns do
+		patterns[i] = P("()", patterns[i], "()")
+	end
+
+	for i = 1, #patterns do
+		patterns[i] = patterns[i]:compile()
+	end
+
+	return patterns
 end
 
 function Cmd.FindNext()
@@ -785,10 +1002,16 @@ function Cmd.FindNext()
 
 	ImmediateMessage("Searching...")
 	
+	-- Get the compiled pattern for the text we're searching for.
+	
+	if not DocumentSet._findpatterns then
+		DocumentSet._findpatterns = compile_patterns(DocumentSet.findtext)
+	end
+	local patterns = DocumentSet._findpatterns
+
 	-- Start at the current cursor position.
 	
 	local cp, cw, co = Document.cp, Document.cw, Document.co
-	local patterns = DocumentSet.findtext
 	if (#patterns == 0) then
 		QueueRedraw()
 		NonmodalMessage("Nothing to search for.")
@@ -801,25 +1024,33 @@ function Cmd.FindNext()
 	
 	while true do
 		local word = Document[cp][cw]
-		local s, e = word.text:find(pattern, co)
+		local s, e = pattern(word, co)
 		
 		if s then
 			-- We got a match! First, though, check to see if the remaining
 			-- words in the pattern match.
 	
-			local endword = cw
+			local ep, ew = cp, cw
 			local pi = 2
 			local found = true
 			while (pi <= #patterns) do
-				endword = endword + 1
+				ew = ew + 1
+				if (ew > #Document[ep]) then
+					ep = ep + 1
+					ew = 1
+					if (ep > #Document) then
+						found = false
+						break
+					end
+				end
 				
-				word = Document[cp][endword]
+				word = Document[ep][ew]
 				if not word then
 					found = false
 					break
 				end
 				
-				_, e = word.text:find(patterns[pi])
+				_, e = patterns[pi](word)
 				if not e then
 					found = false
 					break
@@ -829,10 +1060,10 @@ function Cmd.FindNext()
 			end
 			 
 			if found then
-				Document.cp = cp
-				Document.cw = endword
-				Document.co = e + 1
-				Document.mp = Document.cp
+				Document.cp = ep
+				Document.cw = ew
+				Document.co = e
+				Document.mp = cp
 				Document.mw = cw
 				Document.mo = s
 				NonmodalMessage("Found.")
